#include "itkekADRevolveMattesMutualInformationImageToImageMetric.h"
#include "itkekMacro.h"

#include "itkImageRegionConstIteratorWithIndex.h"
#include "itkImageRegionIterator.h"
#include "itkImageLinearIteratorWithIndex.h"
#include "itkGradientRecursiveGaussianImageFilter.h"
#include "itkImageRandomConstIteratorWithIndex.h"
#include "itkBSplineInterpolateImageFunction.h"
#include "itkCovariantVector.h"
#include "itkImageRandomConstIteratorWithIndex.h"
#include "itkImageRegionConstIterator.h"
#include "itkImageRegionIterator.h"
#include "itkImageIterator.h"
#include "itkBSplineDeformableTransform.h"
#include "vnl/vnl_math.h"
#include "vnl/vnl_sample.h"





namespace itkek
{


// Define static variables

   ADRevolveMattesMutualInformationImageToImageMetric::JointPDFType::Pointer ADRevolveMattesMutualInformationImageToImageMetric::s_JointPDF;
//   ADRevolveMattesMutualInformationImageToImageMetric::JointPDFDerivativesType::Pointer ADRevolveMattesMutualInformationImageToImageMetric::m_JointPDFDerivatives;
   ADRevolveMattesMutualInformationImageToImageMetric::FixedImageSpatialSampleContainer 
          ADRevolveMattesMutualInformationImageToImageMetric::s_FixedImageSamples;
   ADRevolveMattesMutualInformationImageToImageMetric::MarginalPDFType ADRevolveMattesMutualInformationImageToImageMetric::s_FixedImageMarginalPDF;
   ADRevolveMattesMutualInformationImageToImageMetric::MarginalPDFType ADRevolveMattesMutualInformationImageToImageMetric::s_MovingImageMarginalPDF;
   unsigned long ADRevolveMattesMutualInformationImageToImageMetric::s_NumberOfSpatialSamples = 50000;
   unsigned long ADRevolveMattesMutualInformationImageToImageMetric::s_NumberOfParameters;
   unsigned long ADRevolveMattesMutualInformationImageToImageMetric::s_NumberOfHistogramBins = 10;
   double ADRevolveMattesMutualInformationImageToImageMetric::s_MovingImageNormalizedMin;
   double ADRevolveMattesMutualInformationImageToImageMetric::s_FixedImageNormalizedMin;
   double ADRevolveMattesMutualInformationImageToImageMetric::s_MovingImageTrueMin;
   double ADRevolveMattesMutualInformationImageToImageMetric::s_MovingImageTrueMax;
   double ADRevolveMattesMutualInformationImageToImageMetric::s_FixedImageBinSize;
   double ADRevolveMattesMutualInformationImageToImageMetric::s_MovingImageBinSize;
   bool ADRevolveMattesMutualInformationImageToImageMetric::s_UseAllPixels=true;
   ADRevolveMattesMutualInformationImageToImageMetric::CubicBSplineFunctionType::Pointer ADRevolveMattesMutualInformationImageToImageMetric::s_CubicBSplineKernel;
   ADRevolveMattesMutualInformationImageToImageMetric::CubicBSplineDerivativeFunctionType::Pointer ADRevolveMattesMutualInformationImageToImageMetric::s_CubicBSplineDerivativeKernel;
   ADRevolveMattesMutualInformationImageToImageMetric::DerivativeFunctionType::Pointer ADRevolveMattesMutualInformationImageToImageMetric::s_DerivativeCalculator;


   ADRevolveMattesMutualInformationImageToImageMetric::JointPDFTypeAD::Pointer ADRevolveMattesMutualInformationImageToImageMetric::s_JointPDFAD;


/////////////////////////////////////////////////////////
// Setup functions
/////////////////////////////////////////////////////////   




ADRevolveMattesMutualInformationImageToImageMetric::LabelsContainerType 
ADRevolveMattesMutualInformationImageToImageMetric::GetLabels() const
{

   if(s_NumberOfVoxelsPerTimeStep <= 0 || s_FixedImageSamples.size() <= 0)
   {
      itkExceptionMacro(<< "Error: s_NumberOfVoxelsPerTimeStep <= 0 || "
			"s_FixedImageSamples.size() <= 0\n");
   }

   // Setup the labels vector
   LabelsContainerType labels;

   if(m_TransformAD->DoesTransformHaveMoreThanOneTimeStep())
      labels = m_TransformAD->GetLabels();
   else
      labels.push_back(Label(0,0,0));
      

   labels.push_back(Label(1,0,0));
   for(unsigned int i=0;i<s_FixedImageSamples.size(); i++)
   {
      if(i%s_NumberOfVoxelsPerTimeStep==0) 
      { 
	 labels.push_back(Label(1,1,0));
      }
   }
   labels.push_back(Label(1,2,0));



   
//    for(unsigned int i=0;i<labels.size();i++)
//       std::cout << i << ":  " 
// 		<< labels[i].Get(0) << "  " 
// 		<< labels[i].Get(1) << "  " 
// 		<< labels[i].Get(2) << std::endl;
   
   //PRINT(labels.size());


   return labels;

}

#undef FAST_MATRIX_MULTIPLICATION_LABELS


adolc::ObjectiveFunctionBase* 
ADRevolveMattesMutualInformationImageToImageMetric::
Create() const
{ 
   //Pointer p = this->New();
   Self* p = new Self;

   if(!p)
   {
      itkExceptionMacro(<< "Error: Could not allocate new "
	 "ADRevolveMattesMutualInformationImageToImageMetric");
   }

   // Disable reference counting
   p->m_ReferenceCount = 0;

   // Create another transform of the same type as m_TransformAD
   p->m_TransformAD = dynamic_cast<ADRevolveTransformType*>(
      m_TransformAD->CreateAnother().GetPointer());

   if(p->m_TransformAD.IsNull())
   {
      itkExceptionMacro(<< "Error: Could not allocate new "
	 "Transform");
   }


   return p;

}


void
ADRevolveMattesMutualInformationImageToImageMetric
::CopyData(const ObjectiveFunctionBase* rhs)
{
   if(this == rhs)
      return;

   this->Superclass::CopyData(rhs);

   const Self* rhs_cast = static_cast< const Self* >(rhs);

   int numberInts = sizeof(m_Ints) / sizeof(int);

   for(int i=0;i<numberInts;i++)
      m_Ints[i] = rhs_cast->m_Ints[i];

}


/////////////////////////////////////////////////////////
// The next set of functions are called from Function
/////////////////////////////////////////////////////////   




void ADRevolveMattesMutualInformationImageToImageMetric::Function(Label& label)
{
//#ifndef NDEBUG
//   PRINT(label.Get(0) << "  " << label.Get(1));
//     
//   std::cout << "location_cnt: " << location_cnt << std::endl;
//#endif


   JUMPTABLE_1_GOTO_MACRO;

  L0:
   
   if(m_TransformAD->DoesTransformHaveMoreThanOneTimeStep())
      m_TransformAD->Function(label);
   else
      m_TransformAD->SetParametersAD(m_Independents);


   return;
  L1:

   this->MI(label);
   
   return;
}







// The remaining functions are all MI related


/**
 * Initialize
 */
void
ADRevolveMattesMutualInformationImageToImageMetric::
Initialize() throw ( itk::ExceptionObject )
{


  // Cache the number of transformation parameters
   s_NumberOfParameters = m_TransformAD->GetNumberOfParameters();
   //m_NumberOfParameters = 3*m_NumberOfPoints;


  /**
   * Compute the minimum and maximum for the FixedImage over
   * the FixedImageRegion.
   *
   * NB: We can't use StatisticsImageFilter to do this because
   * the filter computes the min/max for the largest possible region
   */
  double fixedImageMin = itk::NumericTraits<double>::max();
  double fixedImageMax = itk::NumericTraits<double>::NonpositiveMin();

  typedef itk::ImageRegionConstIterator<FixedImageType> FixedIteratorType;
  FixedIteratorType fixedImageIterator( 
    m_FixedImage, m_FixedImage->GetLargestPossibleRegion() );

  for ( fixedImageIterator.GoToBegin(); 
        !fixedImageIterator.IsAtEnd(); ++fixedImageIterator )
    {

    double sample = static_cast<double>( fixedImageIterator.Get() );

    if ( sample < fixedImageMin )
      {
      fixedImageMin = sample;
      }

    if ( sample > fixedImageMax )
      {
      fixedImageMax = sample;
      }
    }


  /**
   * Compute the minimum and maximum for the entire moving image
   * in the buffer.
   */
  double movingImageMin = itk::NumericTraits<double>::max();
  double movingImageMax = itk::NumericTraits<double>::NonpositiveMin();

  typedef itk::ImageRegionConstIterator<MovingImageType> MovingIteratorType;
  MovingIteratorType movingImageIterator(
    m_MovingImage, m_MovingImage->GetBufferedRegion() );

  for ( movingImageIterator.GoToBegin(); 
        !movingImageIterator.IsAtEnd(); ++movingImageIterator)
    {
    double sample = static_cast<double>( movingImageIterator.Get() );

    if ( sample < movingImageMin )
      {
      movingImageMin = sample;
      }

    if ( sample > movingImageMax )
      {
      movingImageMax = sample;
      }
    }

  s_MovingImageTrueMin = movingImageMin;
  s_MovingImageTrueMax = movingImageMax;

//  itkDebugMacro( " FixedImageMin: " << fixedImageMin << 
//                 " FixedImageMax: " << fixedImageMax << std::endl );
//  itkDebugMacro( " MovingImageMin: " << movingImageMin << 
//                 " MovingImageMax: " << movingImageMax << std::endl );


  /**
   * Compute binsize for the histograms.
   *
   * The binsize for the image intensities needs to be adjusted so that 
   * we can avoid dealing with boundary conditions using the cubic 
   * spline as the Parzen window.  We do this by increasing the size
   * of the bins so that the joint histogram becomes "padded" at the 
   * borders. Because we are changing the binsize, 
   * we also need to shift the minimum by the padded amount in order to 
   * avoid minimum values filling in our padded region.
   *
   * Note that there can still be non-zero bin values in the padded region,
   * it's just that these bins will never be a central bin for the Parzen
   * window.
   *
   */
  const int padding = 2;  // this will pad by 2 bins

  s_FixedImageBinSize = ( fixedImageMax - fixedImageMin ) /
    static_cast<double>( s_NumberOfHistogramBins - 2 * padding );
  s_FixedImageNormalizedMin = fixedImageMin / s_FixedImageBinSize - 
    static_cast<double>( padding );

  s_MovingImageBinSize = ( movingImageMax - movingImageMin ) /
    static_cast<double>( s_NumberOfHistogramBins - 2 * padding );
  s_MovingImageNormalizedMin = movingImageMin / s_MovingImageBinSize -
    static_cast<double>( padding );


//  itkDebugMacro( "FixedImageNormalizedMin: " << m_FixedImageNormalizedMin );
//  itkDebugMacro( "MovingImageNormalizedMin: " << m_MovingImageNormalizedMin );
//  itkDebugMacro( "FixedImageBinSize: " << m_FixedImageBinSize );
//  itkDebugMacro( "MovingImageBinSize; " << m_MovingImageBinSize );
  

  if( s_UseAllPixels )
    {
       s_NumberOfSpatialSamples = 
	  m_FixedImage->GetLargestPossibleRegion().GetNumberOfPixels();
    }
  
  /**
   * Allocate memory for the fixed image sample container.
   */
  s_FixedImageSamples.resize( s_NumberOfSpatialSamples );

  /**
   * Allocate memory for the marginal PDF and initialize values
   * to zero. The marginal PDFs are stored as std::vector.
   */
  s_FixedImageMarginalPDF.resize( s_NumberOfHistogramBins, 0.0 );
  s_MovingImageMarginalPDF.resize( s_NumberOfHistogramBins, 0.0 );

  /**
   * Allocate memory for the joint PDF and joint PDF derivatives.
   * The joint PDF and joint PDF derivatives are store as itk::Image.
   */
  s_JointPDF = JointPDFType::New();
//  m_JointPDFDerivatives = JointPDFDerivativesType::New();

  // Instantiate a region, index, size
  JointPDFRegionType            jointPDFRegion;
  JointPDFIndexType              jointPDFIndex;
  JointPDFSizeType              jointPDFSize;

//  JointPDFDerivativesRegionType  jointPDFDerivativesRegion;
//  JointPDFDerivativesIndexType  jointPDFDerivativesIndex;
//  JointPDFDerivativesSizeType    jointPDFDerivativesSize;

  // For the joint PDF define a region starting from {0,0} 
  // with size {m_NumberOfHistogramBins, m_NumberOfHistogramBins}.
  // The dimension represents fixed image parzen window index
  // and moving image parzen window index, respectively.
  jointPDFIndex.Fill( 0 ); 
  jointPDFSize.Fill( s_NumberOfHistogramBins ); 

  jointPDFRegion.SetIndex( jointPDFIndex );
  jointPDFRegion.SetSize( jointPDFSize );

  // Set the regions and allocate
  s_JointPDF->SetRegions( jointPDFRegion );
  s_JointPDF->Allocate();

  // For the derivatives of the joint PDF define a region starting from {0,0,0} 
  // with size {m_NumberOfParameters,m_NumberOfHistogramBins, 
  // m_NumberOfHistogramBins}. The dimension represents transform parameters,
  // fixed image parzen window index and moving image parzen window index,
  // respectively. 
//  jointPDFDerivativesIndex.Fill( 0 ); 
  //jointPDFDerivativesSize[0] = m_NumberOfParameters;
  //jointPDFDerivativesSize[1] = m_NumberOfHistogramBins;
  //jointPDFDerivativesSize[2] = m_NumberOfHistogramBins;
//  jointPDFDerivativesSize[0] = m_NumberOfHistogramBins;
//  jointPDFDerivativesSize[1] = m_NumberOfHistogramBins;

//  jointPDFDerivativesRegion.SetIndex( jointPDFDerivativesIndex );
//  jointPDFDerivativesRegion.SetSize( jointPDFDerivativesSize );

  // Set the regions and allocate
//  m_JointPDFDerivatives->SetRegions( jointPDFDerivativesRegion );
//  m_JointPDFDerivatives->Allocate();




  /**
   * Setup the kernels used for the Parzen windows.
   */
  s_CubicBSplineKernel = CubicBSplineFunctionType::New();
  s_CubicBSplineDerivativeKernel = CubicBSplineDerivativeFunctionType::New();    


  if( s_UseAllPixels )
    {
    /** 
     * Take all the pixels within the fixed image region)
     * to create the sample points list.
     */
    this->SampleFullFixedImageDomain( s_FixedImageSamples );
    }
  else
    {
    /** 
     * Uniformly sample the fixed image (within the fixed image region)
     * to create the sample points list.
     */
    this->SampleFixedImageDomain( s_FixedImageSamples );
    }

  /**
   * Pre-compute the fixed image parzen window index for 
   * each point of the fixed image sample points list.
   */
  this->ComputeFixedImageParzenWindowIndices( s_FixedImageSamples );
  
  /**
   * Check if the interpolator is of type BSplineInterpolateImageFunction.
   * If so, we can make use of its EvaluateDerivatives method.
   * Otherwise, we instantiate an external central difference
   * derivative calculator.
   *
   * TODO: Also add it the possibility of using the default gradient
   * provided by the superclass.
   *
   */
/*
  m_InterpolatorIsBSpline = true;

  BSplineInterpolatorType * testPtr = dynamic_cast<BSplineInterpolatorType *>(
    this->m_Interpolator.GetPointer() );
  if ( !testPtr )
    {
    m_InterpolatorIsBSpline = false;
*/

    s_DerivativeCalculator = DerivativeFunctionType::New();
    s_DerivativeCalculator->SetInputImage( m_MovingImage );
    //s_DerivativeCalculator->Print(std::cout);

/*
    m_BSplineInterpolator = NULL;
    itkDebugMacro( "Interpolator is not BSpline" );
    } 
  else
    {
    m_BSplineInterpolator = testPtr;
    m_DerivativeCalculator = NULL;
    itkDebugMacro( "Interpolator is BSpline" );
    }
*/
  /** 
   * Check if the transform is of type BSplineDeformableTransform.
   *
   * If so, several speed up features are implemented.
   * [1] Precomputing the results of bulk transform for each sample point.
   * [2] Precomputing the BSpline weights for each sample point,
   *     to be used later to directly compute the deformation vector
   * [3] Precomputing the indices of the parameters within the 
   *     the support region of each sample point.
   */
/*
  m_TransformIsBSpline = true;

  BSplineTransformType * testPtr2 = dynamic_cast<BSplineTransformType *>(
    this->m_Transform.GetPointer() );
  if( !testPtr2 )
    {
    m_TransformIsBSpline = false;
    m_BSplineTransform = NULL;
    itkDebugMacro( "Transform is not BSplineDeformable" );
    }
  else
    {
    m_BSplineTransform = testPtr2;
    m_NumParametersPerDim = m_BSplineTransform->GetNumberOfParametersPerDimension();
    m_NumBSplineWeights = m_BSplineTransform->GetNumberOfWeights();
    itkDebugMacro( "Transform is BSplineDeformable" );
    }

  if ( m_TransformIsBSpline )
    {
    m_BSplineTransformWeightsArray.SetSize( m_NumberOfSpatialSamples, 
                                            m_NumBSplineWeights );
    m_BSplineTransformIndicesArray.SetSize( m_NumberOfSpatialSamples,
                                            m_NumBSplineWeights );
    m_PreTransformPointsArray.resize( m_NumberOfSpatialSamples );
    m_WithinSupportRegionArray.resize( m_NumberOfSpatialSamples );

    this->PreComputeTransformValues();

    for ( unsigned int j = 0; j < FixedImageDimension; j++ )
      {
      m_ParametersOffset[j] = j * 
        m_BSplineTransform->GetNumberOfParametersPerDimension(); 
      }
    }
*/


    {
   s_JointPDFAD = JointPDFTypeAD::New();
   JointPDFRegionType         jointPDFRegion;
   JointPDFIndexType          jointPDFIndex;
   JointPDFSizeType           jointPDFSize;
   jointPDFIndex.Fill( 0 ); 
   jointPDFSize.Fill( s_NumberOfHistogramBins ); 
   jointPDFRegion.SetIndex( jointPDFIndex );
   jointPDFRegion.SetSize( jointPDFSize );
   s_JointPDFAD->SetRegions( jointPDFRegion );
   s_JointPDFAD->Allocate();
   s_JointPDFAD->FillBuffer( 0.0 );

/*
   // Create a static array of adoubles to pass to the import pointer
   static adouble* jointDataArray = 
      new adouble[s_NumberOfHistogramBins*s_NumberOfHistogramBins];

   s_JointPDFAD->GetPixelContainer()->
      SetImportPointer(jointDataArray,
		       s_NumberOfHistogramBins*s_NumberOfHistogramBins);
*/
    }




    this->Superclass::Initialize();

    PRINT("Finished Initializing Mutual Info.\n\n");
}


/**
 * Uniformly sample the fixed image domain using a random walk
 */
void
ADRevolveMattesMutualInformationImageToImageMetric::
SampleFixedImageDomain( FixedImageSpatialSampleContainer& samples )
{

  // Set up a random interator within the user specified fixed image region.
  typedef itk::ImageRandomConstIteratorWithIndex<FixedImageType> RandomIterator;
  RandomIterator randIter(m_FixedImage,m_FixedImage->GetLargestPossibleRegion());

  randIter.SetNumberOfSamples( s_NumberOfSpatialSamples );
  randIter.GoToBegin();

  FixedImageSpatialSampleContainer::iterator iter;
  FixedImageSpatialSampleContainer::const_iterator end=samples.end();

/*  if( this->m_FixedImageMask )
    {

    typename Superclass::InputPointType inputPoint;

    iter=samples.begin();

    while( iter != end )
      {
      // Get sampled index
      FixedImageIndexType index = randIter.GetIndex();
      // Check if the Index is inside the mask, translate index to point
      this->m_FixedImage->TransformIndexToPhysicalPoint( index, inputPoint );

      // If not inside the mask, ignore the point
      if( !this->m_FixedImageMask->IsInside( inputPoint ) )
        {
        ++randIter; // jump to another random position
        continue;
        }

      // Get sampled fixed image value
      (*iter).FixedImageValue = randIter.Get();
      // Translate index to point
      (*iter).FixedImagePointValue = inputPoint;

      // Jump to random position
      ++randIter;
      ++iter;
      }
    }
  else
*/    {
    for( iter=samples.begin(); iter != end; ++iter )
      {
      // Get sampled index
	 FixedImageType::IndexType index = randIter.GetIndex();
      // Get sampled fixed image value
      (*iter).FixedImageValue = randIter.Get();
      // Translate index to point
      this->m_FixedImage->TransformIndexToPhysicalPoint( index,
                                                  (*iter).FixedImagePointValue );
      // Jump to random position
      ++randIter;

      }
    }
}



/**
 * Sample the fixed image domain using all pixels in the Fixed image region
 */
void
ADRevolveMattesMutualInformationImageToImageMetric::
SampleFullFixedImageDomain( FixedImageSpatialSampleContainer& samples )
{

  // Set up a region interator within the user specified fixed image region.
  typedef itk::ImageRegionConstIteratorWithIndex<FixedImageType> RegionIterator;
  RegionIterator regionIter(m_FixedImage,
			    m_FixedImage->GetLargestPossibleRegion() );

  regionIter.GoToBegin();

  FixedImageSpatialSampleContainer::iterator iter;
  FixedImageSpatialSampleContainer::const_iterator end=samples.end();

/*  if( this->m_FixedImageMask )
    {

    typename Superclass::InputPointType inputPoint;

    iter=samples.begin();

    while( iter != end )
      {
      // Get sampled index
      FixedImageIndexType index = regionIter.GetIndex();
      // Check if the Index is inside the mask, translate index to point
      this->m_FixedImage->TransformIndexToPhysicalPoint( index, inputPoint );

      // If not inside the mask, ignore the point
      if( !this->m_FixedImageMask->IsInside( inputPoint ) )
        {
        ++regionIter; // jump to next pixel
        continue;
        }

      // Get sampled fixed image value
      (*iter).FixedImageValue = regionIter.Get();
      // Translate index to point
      (*iter).FixedImagePointValue = inputPoint;

      // Jump to random position
      ++regionIter;
      ++iter;
      }
    }
  else
*/    {
    for( iter=samples.begin(); iter != end; ++iter )
      {
      // Get sampled index
      FixedImageType::IndexType index = regionIter.GetIndex();
      // Get sampled fixed image value
      (*iter).FixedImageValue = regionIter.Get();
      // Translate index to point
      this->m_FixedImage->TransformIndexToPhysicalPoint( index,
                                                  (*iter).FixedImagePointValue );
      // Jump to random position
      ++regionIter;

      }
    }
}

/**
 * Uniformly sample the fixed image domain using a random walk
 */
void
ADRevolveMattesMutualInformationImageToImageMetric::
ComputeFixedImageParzenWindowIndices( FixedImageSpatialSampleContainer& samples )
{

   FixedImageSpatialSampleContainer::iterator iter;
   FixedImageSpatialSampleContainer::const_iterator end=samples.end();

  for( iter=samples.begin(); iter != end; ++iter )
    {

    // Determine parzen window arguments (see eqn 6 of Mattes paper [2]).  
    double windowTerm =
      static_cast<double>( (*iter).FixedImageValue ) / s_FixedImageBinSize -
        s_FixedImageNormalizedMin;
    unsigned int pindex = static_cast<unsigned int>( floor( windowTerm ) );

    // Make sure the extreme values are in valid bins
    if ( pindex < 2 )
      {
      pindex = 2;
      }
    else if ( pindex > (s_NumberOfHistogramBins - 3) )
      {
      pindex = s_NumberOfHistogramBins - 3;
      }

    (*iter).FixedImageParzenWindowIndex = pindex;

    }

}



void ADRevolveMattesMutualInformationImageToImageMetric::MI(Label& label)
{

   int& j = m_Ints[0];
   int& nFixedImageSamples = m_Ints[1];
   int& nSamples = m_Ints[2];
//   int& c = m_Ints[3];

//   int numberAdoubles = 30 + m_NumberOfHistogramBins +
//      m_NumberOfHistogramBins*m_NumberOfHistogramBins;
//   int ind = m_AdoublesAllocator.GetStartIndex(__LINE__, numberAdoubles);

   static adouble movingImageValueAD = 0.0;
//   adouble& fixedValueAD = m_Adoubles[ind++];
//   adouble& diffAD = m_Adoubles[ind++];
//   adouble& valAD = m_Adoubles[ind++];
   static adouble movingImageParzenWindowTermAD = 0.0;
   static adouble movingImageParzenWindowArgAD = 0.0;
   static adouble absValueAD = 0.0;
   static adouble sqrValueAD = 0.0;
   static adouble jointPDFSumAD = 0.0;
   static adouble sumAD = 0.0;
   static adouble movingImagePDFValueAD = 0.0;
   static adouble jointPDFValueAD = 0.0;
   static adouble pRatioAD = 0.0;
   static adouble temp = 0.0;
   static adouble temp2 = 0.0;

   static ADVectorType inputPointAD(3);
   static ADVectorType mappedPointAD(3);


   static ADVectorType movingImageMarginalPDFAD(s_NumberOfHistogramBins);

/*
   typedef itk::Image<adouble,2> JointPDFTypeAD;
   JointPDFTypeAD::Pointer jointPDFAD;

   jointPDFAD = JointPDFTypeAD::New();
   JointPDFRegionType         jointPDFRegion;
   JointPDFIndexType          jointPDFIndex;
   JointPDFSizeType           jointPDFSize;
   jointPDFIndex.Fill( 0 ); 
   jointPDFSize.Fill( s_NumberOfHistogramBins ); 
   jointPDFRegion.SetIndex( jointPDFIndex );
   jointPDFRegion.SetSize( jointPDFSize );
   jointPDFAD->SetRegions( jointPDFRegion );

   // Create a static array of adoubles to pass to the import pointer
   static adouble* jointDataArray = 
      new adouble[s_NumberOfHistogramBins*s_NumberOfHistogramBins];

   jointPDFAD->GetPixelContainer()->
      SetImportPointer(jointDataArray,
		       s_NumberOfHistogramBins*s_NumberOfHistogramBins);
*/

   JUMPTABLE_2_GOTO_MACRO;
   
  L0:


  // Reset marginal pdf to all zeros.
  // Assumed the size has already been set to NumberOfHistogramBins
  // in Initialize().
  for ( unsigned int jj = 0; jj < s_NumberOfHistogramBins; jj++ )
    {
    s_FixedImageMarginalPDF[jj]  = 0.0;
    movingImageMarginalPDFAD[jj] = 0.0;
    }


  // Reset the joint pdfs to zero
//   m_JointPDF->FillBuffer( 0.0 );
   s_JointPDFAD->FillBuffer( 0.0 );



  // Declare iterators for iteration over the sample container
//  FixedImageSpatialSampleContainer::const_iterator fiter;
//  FixedImageSpatialSampleContainer::const_iterator fend = 
//    m_FixedImageSamples.end();

  nSamples=0;
  nFixedImageSamples=0;


  for(j=0; static_cast<unsigned int>(j) < s_FixedImageSamples.size(); ++j)
     //for ( fiter = m_FixedImageSamples.begin(); fiter != fend; ++fiter )
  {
     
     if(j % s_NumberOfVoxelsPerTimeStep == 0)
     {
	return;
       L1:;
     }

     // Get moving image value
//    MovingImagePointType mappedPoint;
     bool sampleOk;
//    double movingImageValue;

//    this->TransformPoint( nFixedImageSamples, parameters, mappedPoint, 
//                          sampleOk, movingImageValue );
//    this->TransformPointAD( nFixedImageSamples, parametersAD, mappedPointAD, 
//			    sampleOk, movingImageValueAD );
     this->TransformPoint( nFixedImageSamples, mappedPointAD, 
			   sampleOk, movingImageValueAD );

/*
  if(!sampleOk)
  {
  //EVAL(mappedPointAD);
  EVAL(movingImageValueAD);
  EVAL(nFixedImageSamples);
  EVAL(sampleOk);
  EVAL(nFixedImageSamples);
//       EVAL(movingImageValueAD.value()-100.0);
}
*/

     ++nFixedImageSamples;

     if( sampleOk )
     {

	++nSamples; 

	/**
	 * Compute this sample's contribution to the marginal and joint
	 * distributions.
	 *
	 */

	// Determine parzen window arguments (see eqn 6 of Mattes paper [2]).    
	//double movingImageParzenWindowTerm =
	movingImageParzenWindowTermAD =
	   movingImageValueAD / s_MovingImageBinSize - s_MovingImageNormalizedMin;
	//movingImageParzenWindowTermAD =
	//  movingImageValueAD / s_MovingImageBinSize;
	//movingImageParzenWindowTermAD -= s_MovingImageNormalizedMin;

	unsigned int movingImageParzenWindowIndex = 
	   static_cast<unsigned int>( floor(movingImageParzenWindowTermAD.value()));

//      EVAL(movingImageParzenWindowTermAD);

	// Make sure the extreme values are in valid bins
	if ( movingImageParzenWindowIndex < 2 )
        {
	   movingImageParzenWindowIndex = 2;
        }
	else if ( movingImageParzenWindowIndex > (s_NumberOfHistogramBins - 3) )
        {
	   movingImageParzenWindowIndex = s_NumberOfHistogramBins - 3;
        }


	// Since a zero-order BSpline (box car) kernel is used for
	// the fixed image marginal pdf, we need only increment the
	// fixedImageParzenWindowIndex by value of 1.0.
//      m_FixedImageMarginalPDF[(*fiter).FixedImageParzenWindowIndex] += 
//        static_cast<PDFValueType>( 1 );
	s_FixedImageMarginalPDF[s_FixedImageSamples[j].
				FixedImageParzenWindowIndex] += 
	   static_cast<PDFValueType>( 1 );
        
	/**
	 * The region of support of the parzen window determines which bins
	 * of the joint PDF are effected by the pair of image values.
	 * Since we are using a cubic spline for the moving image parzen
	 * window, four bins are affected.  The fixed image parzen window is
	 * a zero-order spline (box car) and thus effects only one bin.
	 *
	 *  The PDF is arranged so that moving image bins corresponds to the 
	 * zero-th (column) dimension and the fixed image bins corresponds
	 * to the first (row) dimension.
	 *
	 */
      
	// Pointer to affected bin to be updated
      
//      JointPDFValueType *pdfPtr = m_JointPDF->GetBufferPointer() +
//     ( (*fiter).FixedImageParzenWindowIndex * m_JointPDF->GetOffsetTable()[1]);

//      JointPDFTypeAD::PixelType* pdfPtr = s_JointPDFAD->GetBufferPointer() +
//      ( (*fiter).FixedImageParzenWindowIndex*s_JointPDFAD->GetOffsetTable()[1] );

	JointPDFTypeAD::PixelType* pdfPtr = s_JointPDFAD->GetBufferPointer() +
	   ( s_FixedImageSamples[j].FixedImageParzenWindowIndex * 
	     s_JointPDFAD->GetOffsetTable()[1] );
 
	// Move the pointer to the first affected bin
	int pdfMovingIndex = static_cast<int>( movingImageParzenWindowIndex ) - 1;
	pdfPtr += pdfMovingIndex;

	for ( ; pdfMovingIndex <= static_cast<int>(movingImageParzenWindowIndex)+2;
	      pdfMovingIndex++, pdfPtr++ )
	{

	   // Update PDF for the current intensity pair
//        double movingImageParzenWindowArg = 
//          static_cast<double>( pdfMovingIndex ) - 
//          static_cast<double>( movingImageParzenWindowTerm );

	   movingImageParzenWindowArgAD = 
	      static_cast<double>( pdfMovingIndex ) - movingImageParzenWindowTermAD;

	   //movingImageParzenWindowArgAD = static_cast<double>( pdfMovingIndex );
	   //movingImageParzenWindowArgAD -= movingImageParzenWindowTermAD;

//        EVAL(movingImageParzenWindowArgAD);

	   // The next two lines are replaced with what follows
//        *(pdfPtr) += static_cast<PDFValueType>( 
//          m_CubicBSplineKernel->Evaluate( movingImageParzenWindowArg ) );
	   {
	      // The following block of code is taken from the file
	      // itkBSplineKernelFunction.h and modified for AD
	   
	      //absValue = vnl_math_abs( u );
	      //sqrValue = vnl_math_sqr( u );
//	   if(movingImageParzenWindowArgAD.value() >= 0.0)
//	      absValueAD = movingImageParzenWindowArgAD;
//	   else
//	      absValueAD = -movingImageParzenWindowArgAD;

	      absValueAD = movingImageParzenWindowArgAD;
	      if(movingImageParzenWindowArgAD.value() < 0.0)
		 absValueAD *= -1.0;

	      sqrValueAD =movingImageParzenWindowArgAD*movingImageParzenWindowArgAD;
//	   sqrValueAD = movingImageParzenWindowArgAD;
//	   sqrValueAD *= sqrValueAD;
	    
//	   adouble tmp=0;
	      if ( absValueAD.value() < 1.0 )
	      {
		 *(pdfPtr) += ( 4.0 - 6.0 * sqrValueAD + 3.0 * sqrValueAD * 
				absValueAD ) * (1.0/6.0);
//	      tmp = ( 4.0 - 6.0 * sqrValueAD + 3.0 * sqrValueAD * 
//		      absValueAD ) / 6.0;

		 /*
		   temp = 4.0;
		   temp2 = -6.0;
		   temp2 *= sqrValueAD;
		   temp += temp2;
		   temp2 = 3.0;
		   temp2 *= sqrValueAD;
		   temp2 *= absValueAD;
		   temp += temp2;
		   temp *= (1.0/6.0);
		   *(pdfPtr) += temp;
		   */
	      }
	      else if ( absValueAD.value() < 2.0 )
	      {
		 *(pdfPtr) += ( 8.0 - 12.0 * absValueAD + 6.0 * sqrValueAD - 
				sqrValueAD * absValueAD ) * (1.0/6.0);
//	      tmp = ( 8.0 - 12.0 * absValueAD + 6.0 * sqrValueAD - 
//		      sqrValueAD * absValueAD ) / 6.0;
	      
		 /*
		   temp = 8.0;
		   temp2 = -12.0;
		   temp2 *= absValueAD;
		   temp += temp2;
		   temp2 = 6.0;
		   temp2 *= sqrValueAD;
		   temp += temp2;
		   temp2 = sqrValueAD;
		   temp2 *= absValueAD;
		   temp -= temp2;
		   temp *= (1.0/6.0);
		   *(pdfPtr) += temp;
		   */


	      }
	      else
	      {
		 //return 0.0;
	      }

//	   EVAL(tmp);
	   }


	}  //end parzen windowing for loop

     } //end if-block check sampleOk

  } // end iterating over fixed image spatial sample container for loop

//  itkDebugMacro( "Ratio of voxels mapping into moving image buffer: " 
//               << nSamples << " / " << m_NumberOfSpatialSamples << std::endl );

  if( static_cast<unsigned int>(nSamples) < s_NumberOfSpatialSamples / 4 )
  {
     itkGenericExceptionMacro("Too many samples map outside moving image buffer: "
			      << nSamples << " / " << s_NumberOfSpatialSamples << std::endl );
  }
  
  return;
  L2:

  /**
   * Normalize the PDFs, compute moving image marginal PDF
   *
   */
  typedef itk::ImageRegionIterator<JointPDFTypeAD> JointPDFIteratorTypeAD;
  JointPDFIteratorTypeAD jointPDFIteratorAD ( s_JointPDFAD, 
					      s_JointPDFAD->GetBufferedRegion() );

  jointPDFIteratorAD.GoToBegin();
  
  // Compute joint PDF normalization factor (to ensure joint PDF sum adds to 1.0)
  jointPDFSumAD = 0.0;

  while( !jointPDFIteratorAD.IsAtEnd() )
    {
    jointPDFSumAD += jointPDFIteratorAD.Get();
//    std::cout << jointPDFIteratorAD.Get().value() << " ";
    ++jointPDFIteratorAD;
    }

  if ( jointPDFSumAD.value() == 0.0 )
    {
    itkGenericExceptionMacro( "Joint PDF summed to zero" );
    }


  // Normalize the PDF bins
  jointPDFIteratorAD.GoToEnd();
  while( !jointPDFIteratorAD.IsAtBegin() )
    {
    --jointPDFIteratorAD;
//    jointPDFIterator.Value() /= static_cast<PDFValueType>( jointPDFSum );
//    jointPDFIteratorAD.Value() /= jointPDFSumAD;
    jointPDFIteratorAD.Value() = jointPDFIteratorAD.Value() / jointPDFSumAD;
    }


  // Normalize the fixed image marginal PDF
  double fixedPDFSum = 0.0;
  for( unsigned int bin = 0; bin < s_NumberOfHistogramBins; bin++ )
    {
    fixedPDFSum += s_FixedImageMarginalPDF[bin];
    }

  if ( fixedPDFSum == 0.0 )
    {
    itkGenericExceptionMacro( "Fixed image marginal PDF summed to zero" );
    }

  for( unsigned int bin=0; bin < s_NumberOfHistogramBins; bin++ )
    {
    s_FixedImageMarginalPDF[bin] /= static_cast<PDFValueType>( fixedPDFSum );
    }


  // Compute moving image marginal PDF by summing over fixed image bins.
  typedef itk::ImageLinearIteratorWithIndex<JointPDFTypeAD> JointPDFLinearIteratorAD;
  JointPDFLinearIteratorAD linearIterAD( 
    s_JointPDFAD, s_JointPDFAD->GetBufferedRegion() );

  linearIterAD.SetDirection( 1 );
  linearIterAD.GoToBegin();
  unsigned int movingIndex = 0;

//  adouble sumAD;

  while( !linearIterAD.IsAtEnd() )
    {
  
       sumAD = 0.0;

    while( !linearIterAD.IsAtEndOfLine() )
      {
      sumAD += linearIterAD.Get();
      ++linearIterAD;
      }

//    m_MovingImageMarginalPDF[movingIndex] = static_cast<PDFValueType>(sumAD);
    movingImageMarginalPDFAD[movingIndex] = sumAD;

    linearIterAD.NextLine();
    ++movingIndex;

    }

  /**
   * Compute the metric by double summation over histogram.
   */

  // Setup pointer to point to the first bin
//  JointPDFValueType * jointPDFPtr = m_JointPDF->GetBufferPointer();
  JointPDFTypeAD::PixelType* jointPDFPtrAD = s_JointPDFAD->GetBufferPointer();

  // Initialze sum to zero
  sumAD = 0.0;

//  adouble movingImagePDFValueAD;
//  adouble jointPDFValueAD;
//  adouble pRatioAD;

  for( unsigned int fixedIndex = 0; fixedIndex < s_NumberOfHistogramBins; 
       ++fixedIndex )
  {

     const double fixedImagePDFValue = s_FixedImageMarginalPDF[fixedIndex];

     for( unsigned int movingIndex = 0; movingIndex < s_NumberOfHistogramBins; 
	  ++movingIndex, jointPDFPtrAD++ )      
     {

//      double movingImagePDFValue = m_MovingImageMarginalPDF[movingIndex];
//      double jointPDFValue = *(jointPDFPtr);
	movingImagePDFValueAD = movingImageMarginalPDFAD[movingIndex];
	jointPDFValueAD = *(jointPDFPtrAD);

	// check for non-zero bin contribution
	if(jointPDFValueAD.value() > 1e-16 && movingImagePDFValueAD.value() >1e-16)
	{

	   //double pRatio = log( jointPDFValue / movingImagePDFValue );
	   pRatioAD = jointPDFValueAD / movingImagePDFValueAD;
	   pRatioAD = log( pRatioAD );

	   if( fixedImagePDFValue > 1e-16)
	   {
//	      sumAD += jointPDFValueAD * ( pRatioAD - log( fixedImagePDFValue ));
	      pRatioAD -= log( fixedImagePDFValue );
	      pRatioAD *= jointPDFValueAD;
	      sumAD += pRatioAD;
	   }

	}  // end if-block to check non-zero bin contribution
     }  // end for-loop over moving index
  }  // end for-loop over fixed index


//  return static_cast<MeasureType>( -1.0 * sum );
//  valAD = -1.0 * sumAD;
  sumAD *= -1.0;



   m_Dependent = sumAD;


   //PRINT(m_Dependent);


}





/**
 * Transform a point from FixedImage domain to MovingImage domain.
 * This function also checks if mapped point is within support region. 
 */

void
ADRevolveMattesMutualInformationImageToImageMetric::
TransformPoint( unsigned int sampleNumber, 
		//MovingImagePointType& mappedPoint,
		ADVectorType& transformedPoint, 
		bool& sampleOk,
		//double& movingImageValue,
		adouble& movingImageValue) const
{

//   int ind = m_TransformPoint_StartIndex;
//   int numberAdoubles = 1;
//   int ind = m_AdoublesAllocator.GetStartIndex(__LINE__, numberAdoubles);

//   static adouble temp;

   // Use generic transform to compute mapped position
   //mappedPoint = this->m_Transform->TransformPoint( 
   //   m_FixedImageSamples[sampleNumber].FixedImagePointValue );
  
   m_TransformAD->TransformPointAD(
      s_FixedImageSamples[sampleNumber].FixedImagePointValue, transformedPoint );



   InputPointType transformedPointReal;
   for (int i=0; i < 3; i++)
      transformedPointReal[i] = transformedPoint[i].value();


   // I modified this so that the sample is always ok. If the transformed
   // point is not inside the image buffer then zero (background) is assumed
   sampleOk = true;

  // Check if mapped point inside image buffer
//  sampleOk = m_Interpolator->IsInsideBuffer( transformedPointReal );
   bool isinside = s_Interpolator->IsInsideBuffer( transformedPointReal );


//  if ( sampleOk )
  if ( isinside )
  {
     const double f = s_Interpolator->Evaluate( transformedPointReal );
     

     // I eliminated this since it does not appear to really be needed.
/*
     if ( f < m_MovingImageTrueMin || 
	  f > m_MovingImageTrueMax )
     {
	// need to throw out this sample as it will not fall into a valid bin
	sampleOk = false;
	return;
     }
*/   

     

     const GradientInterpolatorType::OutputType gradientPoint = 
	//m_GradientInterpolator->Evaluate( transformedPointReal );
	s_DerivativeCalculator->Evaluate( transformedPointReal );

     const double a = gradientPoint[0];
     const double b = gradientPoint[1];
     const double c = gradientPoint[2];
     const double d = f - a*transformedPointReal[0] - 
	b*transformedPointReal[1] - c*transformedPointReal[2];


     movingImageValue = d + transformedPoint[0]*a + 
	transformedPoint[1]*b + transformedPoint[2]*c;
     
/*
     movingImageValue = a;
     movingImageValue *= transformedPoint[0];

     temp = b;
     temp *= transformedPoint[1];
     movingImageValue += temp;

     temp = c;
     temp *= transformedPoint[2];
     movingImageValue += temp;

     movingImageValue += d;
*/
  }
  else
  {
     movingImageValue = 0.0;
     //movingImageValue = m_MovingImageTrueMin;
     //std::cout << sampleNumber << " ";
  }


}




// Method to reinitialize the seed of the random number generator
void
ADRevolveMattesMutualInformationImageToImageMetric::
ReinitializeSeed()
{
  itk::Statistics::MersenneTwisterRandomVariateGenerator::GetInstance()->SetSeed();
}

// Method to reinitialize the seed of the random number generator
void
ADRevolveMattesMutualInformationImageToImageMetric::
ReinitializeSeed(int seed)
{
 itk::Statistics::MersenneTwisterRandomVariateGenerator::GetInstance()->SetSeed(seed);
}



}
