#include "itkekADRevolveMeanSquaresImageToImageMetric.h"

#include "itkImageRegionConstIteratorWithIndex.h"
#include "itkImageRegionIterator.h"
#include "itkImageLinearIteratorWithIndex.h"
#include "itkImageRandomConstIteratorWithIndex.h"
#include "itkImageRandomConstIteratorWithIndex.h"
#include "itkImageRegionConstIterator.h"
#include "itkImageRegionIterator.h"
#include "itkImageIterator.h"

#include "itkekMacro.h"


namespace itkek
{


/////////////////////////////////////////////////////////
// Setup functions
/////////////////////////////////////////////////////////   



ADRevolveMeanSquaresImageToImageMetric::LabelsContainerType 
ADRevolveMeanSquaresImageToImageMetric::GetLabels() const
{
   if(s_NumberOfVoxelsPerTimeStep <= 0 || s_NumberOfVoxelsInImage <= 0)
   {
      itkExceptionMacro(<< "Error: s_NumberOfVoxelsPerTimeStep <= 0 || "
			 "s_NumberOfVoxelsInImage <= 0\n");
   }

   // Setup the labels vector
   LabelsContainerType labels;

   if(m_TransformAD->DoesTransformHaveMoreThanOneTimeStep())
      labels = m_TransformAD->GetLabels();
   else
      labels.push_back(Label(0,0,0));
   

   labels.push_back(Label(1,0,0));
   for(int i=0;i<s_NumberOfVoxelsInImage; i++)
   {
      if(i%s_NumberOfVoxelsPerTimeStep==0) 
      { 
	 labels.push_back(Label(1,1,0));
      }
   }


// #ifndef NDEBUG
//    {
//       for(unsigned int i=0;i<labels.size();i++)
// 	 std::cout << i << ":  " 
// 		   << labels[i].Get(0) << "  " 
// 		   << labels[i].Get(1) << "  " 
// 		   << labels[i].Get(2) << std::endl;
//    }
//#endif

//   PRINT(labels.size());

   return labels;

}






void 
ADRevolveMeanSquaresImageToImageMetric::
Initialize(void) throw ( itk::ExceptionObject )
{


   this->Superclass::Initialize();


}


adolc::ObjectiveFunctionBase* 
ADRevolveMeanSquaresImageToImageMetric::
Create() const
{ 

   Self* p = new Self;

   if(!p)
   {
      itkExceptionMacro(<< "Error: Could not allocate new "
			"ADRevolveMeanSquaresImageToImageMetric");
   }

   // Disable reference counting
   p->m_ReferenceCount = 0;

   // Create another transform of the same type as m_TransformAD
   p->m_TransformAD = dynamic_cast<ADRevolveTransformType*>(
      m_TransformAD->CreateAnother().GetPointer());

   if(p->m_TransformAD.IsNull())
   {
      itkExceptionMacro("Error: Could not allocate new Transform");
   }


   return p;
}


void
ADRevolveMeanSquaresImageToImageMetric
::CopyData(const ObjectiveFunctionBase* rhs)
{
   if(this == rhs)
      return;

   this->Superclass::CopyData(rhs);

   const Self* rhs_cast = static_cast< const Self* >(rhs);

   int numberInts = sizeof(m_Ints) / sizeof(int);

   for(int i=0;i<numberInts;i++)
      m_Ints[i] = rhs_cast->m_Ints[i];
}


/////////////////////////////////////////////////////////
// The next set of functions are called from Function
/////////////////////////////////////////////////////////   




void ADRevolveMeanSquaresImageToImageMetric::Function(Label& label)
{
// #ifndef NDEBUG
//    if(0)
//    {
//       std::cout << label.Get(0) << "  " 
// 		<< label.Get(1) << std::endl;
//      
//       PRINT(location_cnt);
//    }
//#endif


   JUMPTABLE_1_GOTO_MACRO;

  L0:

   if(m_TransformAD->DoesTransformHaveMoreThanOneTimeStep())
      m_TransformAD->Function(label);
   else
      m_TransformAD->SetParametersAD(m_Independents);


   return;
  L1:

   this->SSD(label);
   
   return;
}



void ADRevolveMeanSquaresImageToImageMetric::SSD(Label& label)
{

   int& c = m_Ints[0];
   int& i = m_Ints[1];
   int& j = m_Ints[2];
   int& k = m_Ints[3];
   

   static ADVectorType transformedPoint(3);
   static adouble diff = 0.0;
   static adouble measure = 0.0;

   
   const FixedImageType::RegionType::SizeType s = 
      s_FixedImage->GetLargestPossibleRegion().GetSize();
   
   const int sx = s[0];
   const int sy = s[1];
   const int sz = s[2];

   const FixedImageType::PixelType* fixedImagePtr = 
      s_FixedImage->GetBufferPointer();

   FixedImageType::IndexType index;
   InputPointType inputPoint;
   InputPointType transformedPointReal;


   JUMPTABLE_1_GOTO_MACRO;

  L0:

//    PRINT(s_NumberOfVoxelsInImage);
//    PRINT(s_NumberOfVoxelsPerTimeStep);

   //-------------------------------------------------------------
//   typedef itk::ImageRegionConstIteratorWithIndex<FixedImageType> 
//      FixedIteratorType;
//   FixedIteratorType fixedIterator( s_FixedImage, 
//				    s_FixedImage->GetLargestPossibleRegion() );
   //-------------------------------------------------------------


   //FixedImageType::IndexType index;
   //InputPointType inputPoint;
   //InputPointType transformedPointReal;


   EVAL( if(this->AreWeDoingForwardSweep() == 1) std::cout << "Starting SSD" << std::endl );

   measure = 0.0;
   c = 0;


//   while(!fixedIterator.IsAtEnd())
   //for(j=0; j<m_NumberOfVoxelsInImage; ++j,++fixedIterator)
   for(k=0; k<sz; ++k)
      for(j=0; j<sy; ++j)
	 for(i=0; i<sx; ++i,++c,++fixedImagePtr)
   {

      if(c % s_NumberOfVoxelsPerTimeStep == 0)
      {
	 return;
	L1:;
	 fixedImagePtr += c;
      }




      index[0] = i;
      index[1] = j;
      index[2] = k;
      s_FixedImage->TransformIndexToPhysicalPoint( index, inputPoint);


      m_TransformAD->TransformPointAD( inputPoint, transformedPoint );
      

      for (int ii=0; ii < 3; ii++)
	 transformedPointReal[ii] = transformedPoint[ii].value();

      if( s_Interpolator->IsInsideBuffer( transformedPointReal ) )
      {

//	 const GradientInterpolatorType::OutputType gradient = 
//	    s_GradientInterpolator->Evaluate( transformedPointReal );

	 
	// Get the gradient by NearestNeighboorInterpolation: 
	// which is equivalent to round up the point components.
	typedef  Superclass::OutputPointType OutputPointType;
	typedef  OutputPointType::CoordRepType CoordRepType;
	typedef itk::ContinuousIndex<CoordRepType,3>
	  MovingImageContinuousIndexType;
    
	MovingImageContinuousIndexType tempIndex;
	this->s_MovingImage->TransformPhysicalPointToContinuousIndex( 
	   transformedPointReal, tempIndex );

	MovingImageType::IndexType mappedIndex; 
	for( unsigned int j = 0; j < 3; j++ )
	  {
	    mappedIndex[j] = static_cast<long>( vnl_math_rnd( tempIndex[j] ) );
	  }
    
	const GradientPixelType gradient = 
	   this->s_GradientImage->GetPixel( mappedIndex );


         //const double fixedValue   = fixedIterator.Get();
         const double fixedValue   = *fixedImagePtr;



	 diff = ((s_Interpolator->Evaluate( transformedPointReal ) -
		  gradient[0] * transformedPointReal[0] -
		  gradient[1] * transformedPointReal[1] -
		  gradient[2] * transformedPointReal[2]) - fixedValue) +
	    transformedPoint[0]*gradient[0] + 
	    transformedPoint[1]*gradient[1] + 
	    transformedPoint[2]*gradient[2];

	 measure += diff*diff;

	 
	 
      }
      else
      {
	 
	 //const double movingValue  = 0.0;
	 //const double fixedValue   = fixedIterator.Get();
         const double fixedValue   = *fixedImagePtr;
	 //const double diff = movingValue - fixedValue; 
	 //measureReal += diff * diff; 
	 //diff = -fixedValue;

	 measure += (fixedValue * fixedValue);
	 
      }

      
   }


   //PRINT(measure);
   measure = measure / s_NumberOfVoxelsInImage;
   //PRINT(measure);

   //measure *= m_Weight1; 

   m_Dependent = measure;
   //measure += m_Dependent;
   //PRINT(measure);
   //PRINT(m_Dependent);


}





}
