#!/bin/sh

PATH=/bin:/usr/bin:$PATH
export PATH

# Look for the executables Register and Transform and add their
# directories to the path
REGISTER_PATH=`find $BINARY_DIR -name Register.exe -o -name Register | grep -m 1 '.'`
if [ -z $REGISTER_PATH ]; then
	exit 1
fi
REGISTER_DIR=`dirname $REGISTER_PATH`

TRANSFORM_PATH=`find $BINARY_DIR -name Transform.exe -o -name Transform | grep -m 1 '.'`
if [ -z $TRANSFORM_PATH ]; then
	exit 1
fi
TRANSFORM_DIR=`dirname $TRANSFORM_PATH`

PATH=$REGISTER_DIR:$TRANSFORM_DIR:$PATH
export PATH
#echo $PATH


output_dir=testing_output_`date +%Y-%m-%d-%H%M%S`
mkdir -p $output_dir
log_file=$output_dir/testlog.txt
rm -f $log_file


run_cmd()
{
    cmd=$@
    #echo -e '\n' $cmd '\n' | tee -a $log_file
    #$cmd | tee -a $log_file
    #echo -e '\n' $cmd '\n' >> $log_file 2>&1
    #$cmd >> $log_file 2>&1
    echo -e '\n' $cmd '\n'
    $cmd
    if [ $? != 0 ]; then
		exit 1
    fi

}


#####################################################
#  RIGID


fixedImage=$IRAD_DATA_DIR/sphere_small_shifted_noise.vtk
movingImageFileName=sphere_small_noise.vtk
movingImage=$IRAD_DATA_DIR/$movingImageFileName


# Rigid without AD, SSD
transform_parameters=$output_dir/transform_parameters-i25-m5-t1-z1.txt
transformed_moving_image=$output_dir/transformed-i25-m5-t1-z1-$movingImageFileName
run_cmd Register $fixedImage $movingImage -i25 -m5 -t1 -z1 -p $transform_parameters -e
run_cmd Transform $movingImage $transformed_moving_image $movingImage $transform_parameters

# Rigid with AD no revolve, SSD
transform_parameters=$output_dir/transform_parameters-i25-m3-t1-z1.txt
transformed_moving_image=$output_dir/transformed-i25-m3-t1-z1-$movingImageFileName
run_cmd Register $fixedImage $movingImage -i25 -m3 -t1 -z1 -p $transform_parameters -e
run_cmd Transform $movingImage $transformed_moving_image $movingImage $transform_parameters

# Rigid with AD and revolve, SSD
transform_parameters=$output_dir/transform_parameters-i25-m1-t1-z1.txt
transformed_moving_image=$output_dir/transformed-i25-m1-t1-z1-$movingImageFileName
run_cmd Register $fixedImage $movingImage -i25 -m1 -t1 -z1 -p $transform_parameters -e
run_cmd Transform $movingImage $transformed_moving_image $movingImage $transform_parameters

# Rigid without AD, MI
transform_parameters=$output_dir/transform_parameters-i25-m6-t1-z1.txt
transformed_moving_image=$output_dir/transformed-i25-m6-t1-z1-$movingImageFileName
run_cmd Register $fixedImage $movingImage -i25 -m6 -t1 -z1 -p $transform_parameters -e
run_cmd Transform $movingImage $transformed_moving_image $movingImage $transform_parameters

# Rigid with AD and revolve, MI
transform_parameters=$output_dir/transform_parameters-i25-m2-t1-z1.txt
transformed_moving_image=$output_dir/transformed-i25-m2-t1-z1-$movingImageFileName
run_cmd Register $fixedImage $movingImage -i25 -m2 -t1 -z1 -p $transform_parameters -e
run_cmd Transform $movingImage $transformed_moving_image $movingImage $transform_parameters



#####################################################
# AFFINE

fixedImage=$IRAD_DATA_DIR/sphere_small_shifted_noise.vtk
movingImageFileName=sphere_noise.vtk
movingImage=$IRAD_DATA_DIR/$movingImageFileName


# Affine without AD, SSD
transform_parameters=$output_dir/transform_parameters-i25-m5-t2-z1.txt
transformed_moving_image=$output_dir/transformed-i25-m5-t2-z1-$movingImageFileName
run_cmd Register $fixedImage $movingImage -i25 -m5 -t2 -z1 -p $transform_parameters -e
run_cmd Transform $movingImage $transformed_moving_image $movingImage $transform_parameters

# Affine with AD no revolve, SSD
transform_parameters=$output_dir/transform_parameters-i25-m3-t2-z1.txt
transformed_moving_image=$output_dir/transformed-i25-m3-t2-z1-$movingImageFileName
run_cmd Register $fixedImage $movingImage -i25 -m3 -t2 -z1 -p $transform_parameters -e
run_cmd Transform $movingImage $transformed_moving_image $movingImage $transform_parameters

# Affine with AD and revolve, SSD
transform_parameters=$output_dir/transform_parameters-i25-m1-t2-z1.txt
transformed_moving_image=$output_dir/transformed-i25-m1-t2-z1-$movingImageFileName
run_cmd Register $fixedImage $movingImage -i25 -m1 -t2 -z1 -p $transform_parameters -e
run_cmd Transform $movingImage $transformed_moving_image $movingImage $transform_parameters

# Affine without AD, MI
transform_parameters=$output_dir/transform_parameters-i25-m6-t2-z1.txt
transformed_moving_image=$output_dir/transformed-i25-m6-t2-z1-$movingImageFileName
run_cmd Register $fixedImage $movingImage -i25 -m6 -t2 -z1 -p $transform_parameters -e
run_cmd Transform $movingImage $transformed_moving_image $movingImage $transform_parameters

# Affine with AD and revolve, MI
transform_parameters=$output_dir/transform_parameters-i25-m2-t2-z1.txt
transformed_moving_image=$output_dir/transformed-i25-m2-t2-z1-$movingImageFileName
run_cmd Register $fixedImage $movingImage -i25 -m2 -t2 -z1 -p $transform_parameters -e
run_cmd Transform $movingImage $transformed_moving_image $movingImage $transform_parameters



#####################################################
# B-SPLINE

gridImageFileName=grid.vtk
gridImage=$IRAD_DATA_DIR/$gridImageFileName

fixedImage=$IRAD_DATA_DIR/sphere.vtk
movingImageFileName=planet.vtk
movingImage=$IRAD_DATA_DIR/$movingImageFileName


# BSpline without AD, SSD
transform_parameters=$output_dir/transform_parameters-i25-m5-t3-z2.txt
transformed_moving_image=$output_dir/transformed-i25-m5-t3-z2-$movingImageFileName
transformed_grid_image=$output_dir/transformed-i25-m5-t3-z2-$gridImageFileName
run_cmd Register $fixedImage $movingImage -i25 -m5 -t3 -z2 -p $transform_parameters -e
run_cmd Transform $movingImage $transformed_moving_image $movingImage $transform_parameters
run_cmd Transform $gridImage $transformed_grid_image $gridImage $transform_parameters

# BSpline with AD no revolve, SSD
transform_parameters=$output_dir/transform_parameters-i25-m3-t3-z2.txt
transformed_moving_image=$output_dir/transformed-i25-m3-t3-z2-$movingImageFileName
transformed_grid_image=$output_dir/transformed-i25-m3-t3-z2-$gridImageFileName
run_cmd Register $fixedImage $movingImage -i25 -m3 -t3 -z2 -p $transform_parameters -e
run_cmd Transform $movingImage $transformed_moving_image $movingImage $transform_parameters
run_cmd Transform $gridImage $transformed_grid_image $gridImage $transform_parameters

# BSpline with AD and revolve, SSD
transform_parameters=$output_dir/transform_parameters-i25-m1-t3-z2.txt
transformed_moving_image=$output_dir/transformed-i25-m1-t3-z2-$movingImageFileName
transformed_grid_image=$output_dir/transformed-i25-m1-t3-z2-$gridImageFileName
run_cmd Register $fixedImage $movingImage -i25 -m1 -t3 -z2 -p $transform_parameters -e
run_cmd Transform $movingImage $transformed_moving_image $movingImage $transform_parameters
run_cmd Transform $gridImage $transformed_grid_image $gridImage $transform_parameters

# BSpline without AD, MI
#transform_parameters=$output_dir/transform_parameters-i25-m6-t3-z2.txt
#transformed_moving_image=$output_dir/transformed-i25-m6-t3-z2-$movingImageFileName
#transformed_grid_image=$output_dir/transformed-i25-m6-t3-z2-$gridImageFileName
#run_cmd Register $fixedImage $movingImage -i25 -m6 -t3 -z2 -p $transform_parameters -e
#run_cmd Transform $movingImage $transformed_moving_image $movingImage $transform_parameters
#run_cmd Transform $gridImage $transformed_grid_image $gridImage $transform_parameters

# BSpline with AD and revolve, MI
#transform_parameters=$output_dir/transform_parameters-i25-m2-t3-z2.txt
#transformed_moving_image=$output_dir/transformed-i25-m2-t3-z2-$movingImageFileName
#transformed_grid_image=$output_dir/transformed-i25-m2-t3-z2-$gridImageFileName
#run_cmd Register $fixedImage $movingImage -i25 -m2 -t3 -z2 -p $transform_parameters -e
#run_cmd Transform $movingImage $transformed_moving_image $movingImage $transform_parameters
#run_cmd Transform $gridImage $transformed_grid_image $gridImage $transform_parameters


#####################################################
# TPS

controlPointsFile=$IRAD_DATA_DIR/control_points_file.txt

# TPS with AD and revolve, SSD
transform_parameters=$output_dir/transform_parameters-i10-m1-t4-z2.txt
transformed_moving_image=$output_dir/transformed-i10-m1-t4-z2-$movingImageFileName
transformed_grid_image=$output_dir/transformed-i10-m1-t4-z2-$gridImageFileName
run_cmd Register $fixedImage $movingImage -i10 -m1 -t4 -z2 -c $controlPointsFile -p $transform_parameters -e
run_cmd Transform $movingImage $transformed_moving_image $movingImage $transform_parameters
run_cmd Transform $gridImage $transformed_grid_image $gridImage $transform_parameters

# TPS with AD and revolve, MI
#transform_parameters=$output_dir/transform_parameters-i10-m2-t4-z2.txt
#transformed_moving_image=$output_dir/transformed-i10-m2-t4-z2-$movingImageFileName
#transformed_grid_image=$output_dir/transformed-i10-m2-t4-z2-$gridImageFileName
#run_cmd Register $fixedImage $movingImage -i10 -m2 -t4 -z2 -c $controlPointsFile -p $transform_parameters -e
#run_cmd Transform $movingImage $transformed_moving_image $movingImage $transform_parameters
#run_cmd Transform $gridImage $transformed_grid_image $gridImage $transform_parameters
