#ifndef _itkekADRevolveBSplineDeformableTransform_txx
#define _itkekADRevolveBSplineDeformableTransform_txx

#include "itkekADRevolveBSplineDeformableTransform.h"
#include "itkContinuousIndex.h"
#include "itkImageRegionIterator.h"
#include "itkImageRegionConstIterator.h"
#include "itkImageRegionConstIteratorWithIndex.h"
#include "itkIdentityTransform.h"
#include "itkAffineTransform.h"
#include <fstream>

namespace itkek
{


// define static stuff

template<class TScalarType, unsigned int NDimensions, unsigned int VSplineOrder>
const adouble* 
ADRevolveBSplineDeformableTransform<TScalarType, NDimensions,VSplineOrder>
::s_InputParametersPointerAD = 0;

template<class TScalarType, unsigned int NDimensions, unsigned int VSplineOrder>
typename ADRevolveBSplineDeformableTransform<TScalarType, NDimensions,VSplineOrder>
::BulkTransformPointer
ADRevolveBSplineDeformableTransform<TScalarType, NDimensions,VSplineOrder>
::s_BulkTransform;
  
template<class TScalarType, unsigned int NDimensions, unsigned int VSplineOrder>
typename ADRevolveBSplineDeformableTransform<TScalarType, NDimensions,VSplineOrder>
::ImagePointer
ADRevolveBSplineDeformableTransform<TScalarType, NDimensions,VSplineOrder>
::s_CoefficientImage[NDimensions];
  
template<class TScalarType, unsigned int NDimensions, unsigned int VSplineOrder>
typename ADRevolveBSplineDeformableTransform<TScalarType, NDimensions,VSplineOrder>
::SpacingType
ADRevolveBSplineDeformableTransform<TScalarType, NDimensions,VSplineOrder>
::s_GridSpacing;
  
template<class TScalarType, unsigned int NDimensions, unsigned int VSplineOrder>
typename ADRevolveBSplineDeformableTransform<TScalarType, NDimensions,VSplineOrder>
::OriginType
ADRevolveBSplineDeformableTransform<TScalarType, NDimensions,VSplineOrder>
::s_GridOrigin;

template<class TScalarType, unsigned int NDimensions, unsigned int VSplineOrder>
typename ADRevolveBSplineDeformableTransform<TScalarType, NDimensions,VSplineOrder>
::RegionType
ADRevolveBSplineDeformableTransform<TScalarType, NDimensions,VSplineOrder>
::s_GridRegion;

template<class TScalarType, unsigned int NDimensions, unsigned int VSplineOrder>
typename ADRevolveBSplineDeformableTransform<TScalarType, NDimensions,VSplineOrder>
::ParametersType
ADRevolveBSplineDeformableTransform<TScalarType, NDimensions,VSplineOrder>
::s_DummyParameters;

template<class TScalarType, unsigned int NDimensions, unsigned int VSplineOrder>
typename ADRevolveBSplineDeformableTransform<TScalarType, NDimensions,VSplineOrder>
::RegionType
ADRevolveBSplineDeformableTransform<TScalarType, NDimensions,VSplineOrder>
::s_ValidRegion;

template<class TScalarType, unsigned int NDimensions, unsigned int VSplineOrder>
typename ADRevolveBSplineDeformableTransform<TScalarType, NDimensions,VSplineOrder>
::IndexType
ADRevolveBSplineDeformableTransform<TScalarType, NDimensions,VSplineOrder>
::s_ValidRegionLast;



// Constructor with default arguments
template<class TScalarType, unsigned int NDimensions, unsigned int VSplineOrder>
ADRevolveBSplineDeformableTransform<TScalarType, NDimensions,VSplineOrder>
::ADRevolveBSplineDeformableTransform():Superclass(SpaceDimension,0)
{

  // Instantiate a weights function
  m_WeightsFunction = WeightsFunctionType::New();
  m_SupportSize = m_WeightsFunction->GetSupportSize();

  // Instantiate an identity transform
  typedef itk::IdentityTransform<ScalarType,SpaceDimension> IdentityTransformType;
  typename IdentityTransformType::Pointer id = IdentityTransformType::New();
  m_BulkTransform = id;

  // Default grid size is zero
  typename RegionType::SizeType size;
  typename RegionType::IndexType index;
  size.Fill( 0 );
  index.Fill( 0 );
  m_GridRegion.SetSize( size );
  m_GridRegion.SetIndex( index );

  m_GridOrigin.Fill( 0.0 );  // default origin is all zeros
  m_GridSpacing.Fill( 1.0 ); // default spacing is all ones

  m_InputParametersPointer = NULL;

  // Initialize coeffient images
  for ( unsigned int j = 0; j < SpaceDimension; j++ )
    {
    m_WrappedImage[j] = ImageType::New();
    m_WrappedImage[j]->SetRegions( m_GridRegion );
    m_WrappedImage[j]->SetOrigin( m_GridOrigin.GetDataPointer() );
    m_WrappedImage[j]->SetSpacing( m_GridSpacing.GetDataPointer() );
    m_CoefficientImage[j] = NULL;
    }

  // Setup variables for computing interpolation
  m_Offset = SplineOrder / 2;
  if ( SplineOrder % 2 ) 
    {
    m_SplineOrderOdd = true;
    }
  else
    {
    m_SplineOrderOdd = false;
    }
  m_ValidRegion = m_GridRegion;

  // Initialize jacobian images
  for ( unsigned int j = 0; j < SpaceDimension; j++ )
    {
    m_JacobianImage[j] = ImageType::New();
    m_JacobianImage[j]->SetRegions( m_GridRegion );
    m_JacobianImage[j]->SetOrigin( m_GridOrigin.GetDataPointer() );
    m_JacobianImage[j]->SetSpacing( m_GridSpacing.GetDataPointer() );
    }

  m_LastJacobianIndex = m_ValidRegion.GetIndex();


// EGK Initialize the static coefficient images if they have not already been
// done so.
  if(s_CoefficientImage[0].IsNull())
  {
     for ( unsigned int j = 0; j < SpaceDimension; j++ )
     {
	 s_CoefficientImage[j] = ImageType::New();
     }
     s_ValidRegion = m_ValidRegion;
     s_ValidRegionLast = m_ValidRegionLast;
  }

}
    

// Destructor
template<class TScalarType, unsigned int NDimensions, unsigned int VSplineOrder>
ADRevolveBSplineDeformableTransform<TScalarType, NDimensions,VSplineOrder>
::~ADRevolveBSplineDeformableTransform()
{

}



// Get the number of parameters
template<class TScalarType, unsigned int NDimensions, unsigned int VSplineOrder>
unsigned int
ADRevolveBSplineDeformableTransform<TScalarType, NDimensions,VSplineOrder>
::GetNumberOfParameters(void) const
{

  // The number of parameters equal SpaceDimension * number of
  // of pixels in the grid region.
  return ( static_cast<unsigned int>( SpaceDimension ) *
           static_cast<unsigned int>( m_GridRegion.GetNumberOfPixels() ) );

}


// Get the number of parameters per dimension
template<class TScalarType, unsigned int NDimensions, unsigned int VSplineOrder>
unsigned int
ADRevolveBSplineDeformableTransform<TScalarType, NDimensions,VSplineOrder>
::GetNumberOfParametersPerDimension(void) const
{
  // The number of parameters per dimension equal number of
  // of pixels in the grid region.
  return ( static_cast<unsigned int>( m_GridRegion.GetNumberOfPixels() ) );

}


// Set the grid region
template<class TScalarType, unsigned int NDimensions, unsigned int VSplineOrder>
void
ADRevolveBSplineDeformableTransform<TScalarType, NDimensions,VSplineOrder>
::SetGridRegion( const RegionType& region )
{
  if ( m_GridRegion != region )
    {

    m_GridRegion = region;

    // set regions for each coefficient and jacobian image
    for ( unsigned int j = 0; j < SpaceDimension; j++ )
      {
      m_WrappedImage[j]->SetRegions( m_GridRegion );
      m_JacobianImage[j]->SetRegions( m_GridRegion );
      }

    // Set the valid region
    // If the grid spans the interval [start,last].
    // The valid interval for evaluation is [start+offset,last-offset] when spline order is even.
    // The valid interval for evaluation is [start+offset,last-offset) when spline order is odd
    // Where offset = floor( spline / 2 ).
    // Note that the last pixel is not included in the valid region with odd spline orders.
    typename RegionType::SizeType size = m_GridRegion.GetSize();
    typename RegionType::IndexType index = m_GridRegion.GetIndex();
    for ( unsigned int j = 0; j < SpaceDimension; j++ )
      {
      index[j] += 
        static_cast< typename RegionType::IndexValueType >( m_Offset );
      size[j] -= 
        static_cast< typename RegionType::SizeValueType> ( 2 * m_Offset );
      m_ValidRegionLast[j] = index[j] +
        static_cast< typename RegionType::IndexValueType >( size[j] ) - 1;
      }
    m_ValidRegion.SetSize( size );
    m_ValidRegion.SetIndex( index );

    this->Modified();


    s_GridRegion = region; //EGK
    for ( unsigned int j = 0; j < SpaceDimension; j++ )
      {
      s_CoefficientImage[j]->SetRegions( s_GridRegion );
      }
    s_ValidRegion = m_ValidRegion;
    s_ValidRegionLast = m_ValidRegionLast;

    }

}


// Set the grid spacing
template<class TScalarType, unsigned int NDimensions, unsigned int VSplineOrder>
void
ADRevolveBSplineDeformableTransform<TScalarType, NDimensions,VSplineOrder>
::SetGridSpacing( const SpacingType& spacing )
{
  if ( m_GridSpacing != spacing )
    {
    m_GridSpacing = spacing;

    // set spacing for each coefficient and jacobian image
    for ( unsigned int j = 0; j < SpaceDimension; j++ )
      {
      m_WrappedImage[j]->SetSpacing( m_GridSpacing.GetDataPointer() );
      m_JacobianImage[j]->SetSpacing( m_GridSpacing.GetDataPointer() );
      }

    this->Modified();

    s_GridSpacing = spacing; //EGK
    for ( unsigned int j = 0; j < SpaceDimension; j++ )
      {
      s_CoefficientImage[j]->SetSpacing( s_GridSpacing.GetDataPointer() );
      }

    }

}


// Set the grid origin
template<class TScalarType, unsigned int NDimensions, unsigned int VSplineOrder>
void
ADRevolveBSplineDeformableTransform<TScalarType, NDimensions,VSplineOrder>
::SetGridOrigin( const OriginType& origin )
{
  if ( m_GridOrigin != origin )
    {
    m_GridOrigin = origin;

    // set spacing for each coefficient and jacobianimage
    for ( unsigned int j = 0; j < SpaceDimension; j++ )
      {
      m_WrappedImage[j]->SetOrigin( m_GridOrigin.GetDataPointer() );
      m_JacobianImage[j]->SetOrigin( m_GridOrigin.GetDataPointer() );
      }

    this->Modified();

    s_GridOrigin = origin;
    for ( unsigned int j = 0; j < SpaceDimension; j++ )
      {
      s_CoefficientImage[j]->SetOrigin( s_GridOrigin.GetDataPointer() );
      }

    }

}


// Set the parameters
template<class TScalarType, unsigned int NDimensions, unsigned int VSplineOrder>
void
ADRevolveBSplineDeformableTransform<TScalarType, NDimensions,VSplineOrder>
::SetIdentity()
{
   if( m_InputParametersPointer )
     {
     ParametersType * parameters = const_cast<ParametersType *>( m_InputParametersPointer );
     parameters->Fill( 0.0 );
     this->Modified();
     }
}


// Set the parameters
template<class TScalarType, unsigned int NDimensions, unsigned int VSplineOrder>
void
ADRevolveBSplineDeformableTransform<TScalarType, NDimensions,VSplineOrder>
::SetParameters( const ParametersType & parameters )
{

  // check if the number of parameters match the
  // expected number of parameters
  if ( parameters.Size() != this->GetNumberOfParameters() )
    {
    itkExceptionMacro(<<"Mismatched between parameters size " << parameters.size() 
                      << " and region size " << m_GridRegion.GetNumberOfPixels() );
    }

  // Keep a reference to the input parameters
  m_InputParametersPointer = &parameters;

  /**
   * Wrap flat parameters array into SpaceDimension number of ITK images
   * NOTE: For efficiency, parameters are not copied locally. The parameters
   * are assumed to be maintained by the caller.
   */
  //PixelType * dataPointer = static_cast<PixelType *>( parameters.data_block() );
  PixelType * dataPointer = const_cast<PixelType *>(( parameters.data_block() ));
  unsigned int numberOfPixels = m_GridRegion.GetNumberOfPixels();

  for ( unsigned int j = 0; j < SpaceDimension; j++ )
    {
    m_WrappedImage[j]->GetPixelContainer()->
      SetImportPointer( dataPointer, numberOfPixels );
    dataPointer += numberOfPixels;
    m_CoefficientImage[j] = m_WrappedImage[j];
    }

  /**
   * Allocate memory for Jacobian and wrap into SpaceDimension number
   * of ITK images
   */
  this->m_Jacobian.set_size( SpaceDimension, this->GetNumberOfParameters() );
  this->m_Jacobian.Fill( itk::NumericTraits<JacobianPixelType>::Zero );
  m_LastJacobianIndex = m_ValidRegion.GetIndex();
  JacobianPixelType * jacobianDataPointer = this->m_Jacobian.data_block();

  for ( unsigned int j = 0; j < SpaceDimension; j++ )
    {
    m_JacobianImage[j]->GetPixelContainer()->
      SetImportPointer( jacobianDataPointer, numberOfPixels );
    jacobianDataPointer += this->GetNumberOfParameters() + numberOfPixels;
    }
  // Modified is always called since we just have a pointer to the
  // parameters and cannot know if the parameters have changed.
  this->Modified();
}


// Set the parameters
// In order to make this class compatible with AD, these 2 function are
// really the only functions which we need to define.
// Note: the function SetParameters in the base class must also be called
// in addition to this function.
template< class TScalarType, unsigned int NDimensions, unsigned int VSplineOrder>
void
ADRevolveBSplineDeformableTransform< TScalarType, NDimensions, VSplineOrder>
::SetParametersAD( const adouble* parameters )
{

  if ( !parameters )
    {
    itkExceptionMacro(<<"parameters is null" );
    }

  s_InputParametersPointerAD = parameters;

  unsigned int numberOfParameters = 
     SpaceDimension*s_GridRegion.GetNumberOfPixels();

  if ( numberOfParameters == 0)
    {
       itkExceptionMacro(<<"number of parameters is zero" );
    }


   if(numberOfParameters != s_DummyParameters.GetSize())
   {
      s_DummyParameters.SetSize(numberOfParameters);


      /**
       * Wrap flat parameters array into SpaceDimension number of ITK images
       * NOTE: For efficiency, parameters are not copied locally. The parameters
       * are assumed to be maintained by the caller.
       */
      //PixelType * dataPointer = static_cast<PixelType *>( parameters.data_block() );
      PixelType * dataPointer = const_cast<PixelType *>(( s_DummyParameters.data_block() ));
      unsigned int numberOfPixels = s_GridRegion.GetNumberOfPixels();

      for ( unsigned int j = 0; j < SpaceDimension; j++ )
      {
	  s_CoefficientImage[j]->GetPixelContainer()->
	     SetImportPointer( dataPointer, numberOfPixels );
	  dataPointer += numberOfPixels;
      }

   }

}



// Get the parameters
template<class TScalarType, unsigned int NDimensions, unsigned int VSplineOrder>
const 
typename ADRevolveBSplineDeformableTransform<TScalarType, NDimensions,VSplineOrder>
::ParametersType &
ADRevolveBSplineDeformableTransform<TScalarType, NDimensions,VSplineOrder>
::GetParameters( void ) const
{
  /** NOTE: For efficiency, this class does not keep a copy of the parameters - 
   * it just keeps pointer to input parameters. 
   */
  return (*m_InputParametersPointer);
}

  
// Set the B-Spline coefficients using input images
template<class TScalarType, unsigned int NDimensions, unsigned int VSplineOrder>
void 
ADRevolveBSplineDeformableTransform<TScalarType, NDimensions,VSplineOrder>
::SetCoefficientImage( ImagePointer images[] )
{
  if ( images[0] )
    {
    this->SetGridRegion( images[0]->GetBufferedRegion() );
    this->SetGridSpacing( images[0]->GetSpacing() );
    this->SetGridOrigin( images[0]->GetOrigin() );

    for( unsigned int j = 0; j < SpaceDimension; j++ )
      {
      m_CoefficientImage[j] = images[j];
      }
    }
}  

// Print self
template<class TScalarType, unsigned int NDimensions, unsigned int VSplineOrder>
void
ADRevolveBSplineDeformableTransform<TScalarType, NDimensions,VSplineOrder>
::PrintSelf(std::ostream &os, itk::Indent indent) const
{

  unsigned int j;

  this->Superclass::PrintSelf(os,indent);

  os << indent << "GridRegion: " << m_GridRegion << std::endl;
  os << indent << "GridOrigin: " << m_GridOrigin << std::endl;
  os << indent << "GridSpacing: " << m_GridSpacing << std::endl;

  os << indent << "CoefficientImage: [ ";
  for ( j = 0; j < SpaceDimension - 1; j++ )
    {
    os << m_CoefficientImage[j].GetPointer() << ", ";
    }
  os << m_CoefficientImage[j].GetPointer() << " ]" << std::endl;

  os << indent << "WrappedImage: [ ";
  for ( j = 0; j < SpaceDimension - 1; j++ )
    {
    os << m_WrappedImage[j].GetPointer() << ", ";
    }
  os << m_WrappedImage[j].GetPointer() << " ]" << std::endl;
 
  os << indent << "InputParametersPointer: " 
     << m_InputParametersPointer << std::endl;
  os << indent << "ValidRegion: " << m_ValidRegion << std::endl;
  os << indent << "LastJacobianIndex: " << m_LastJacobianIndex << std::endl;
  os << indent << "BulkTransform: ";
  os << m_BulkTransform.GetPointer() << std::endl;
  os << indent << "WeightsFunction: ";
  os << m_WeightsFunction.GetPointer() << std::endl;

  if ( m_BulkTransform )
    {
    os << indent << "BulkTransformType: " 
       << m_BulkTransform->GetNameOfClass() << std::endl;
    }
     
}

// Transform a point
template<class TScalarType, unsigned int NDimensions, unsigned int VSplineOrder>
bool 
ADRevolveBSplineDeformableTransform<TScalarType, NDimensions,VSplineOrder>
::InsideValidRegion( 
  const ContinuousIndexType& index ) const
{
  bool inside = true;

  if ( !m_ValidRegion.IsInside( index ) )
    {
    inside = false;
    }

  if ( inside && m_SplineOrderOdd )
    {
    typedef typename ContinuousIndexType::ValueType ValueType;
    for( unsigned int j = 0; j < SpaceDimension; j++ )
      {
      if ( index[j] >= static_cast<ValueType>( m_ValidRegionLast[j] ) )
        { 
        inside = false;
        break;
        }
      }
    }

  return inside;
}

// Transform a point
template<class TScalarType, unsigned int NDimensions, unsigned int VSplineOrder>
bool 
ADRevolveBSplineDeformableTransform<TScalarType, NDimensions,VSplineOrder>
::InsideValidRegionAD( 
  const ContinuousIndexType& index ) const
{
  bool inside = true;

  if ( !s_ValidRegion.IsInside( index ) )
    {
    inside = false;
    }

  if ( inside && m_SplineOrderOdd )
    {
    typedef typename ContinuousIndexType::ValueType ValueType;
    for( unsigned int j = 0; j < SpaceDimension; j++ )
      {
      if ( index[j] >= static_cast<ValueType>( s_ValidRegionLast[j] ) )
        { 
        inside = false;
        break;
        }
      }
    }

  return inside;
}


/*
// Transform a point
template<class TScalarType, unsigned int NDimensions, unsigned int VSplineOrder>
void 
ADRevolveBSplineDeformableTransform<TScalarType, NDimensions,VSplineOrder>
::TransformPoint( 
  const InputPointType & point, 
  OutputPointType & outputPoint, 
  WeightsType & weights, 
  ParameterIndexArrayType & indices,
  bool& inside ) const
{

  unsigned int j;
  IndexType supportIndex;

  InputPointType transformedPoint;
  if ( m_BulkTransform )
    {
    transformedPoint = m_BulkTransform->TransformPoint( point );
    }
  else
    {
    transformedPoint = point;
    }

  if ( m_CoefficientImage[0] )
    {

    ContinuousIndexType index;
    for ( j = 0; j < SpaceDimension; j++ )
      {
      index[j] = ( point[j] - m_GridOrigin[j] ) / m_GridSpacing[j];
      }

    // NOTE: if the support region does not lie totally within the grid
    // we assume zero displacement and return the input point
    inside = this->InsideValidRegion( index );
    if ( !inside )
      {
      outputPoint = transformedPoint;
      return;
      }

    // Compute interpolation weights
    m_WeightsFunction->Evaluate( index, weights, supportIndex );

    // For each dimension, correlate coefficient with weights
    RegionType supportRegion;
    supportRegion.SetSize( m_SupportSize );
    supportRegion.SetIndex( supportIndex );

    outputPoint.Fill( itk::NumericTraits<ScalarType>::Zero );

    typedef itk::ImageRegionConstIterator<ImageType> IteratorType;
    IteratorType m_Iterator[ SpaceDimension ];
    unsigned long counter = 0;
    const PixelType * basePointer = m_CoefficientImage[0]->GetBufferPointer();

    for ( j = 0; j < SpaceDimension; j++ )
      {
      m_Iterator[j] = IteratorType( m_CoefficientImage[j], supportRegion );
      }

    while ( ! m_Iterator[0].IsAtEnd() )
      {

      // multiply weight with coefficient
      for ( j = 0; j < SpaceDimension; j++ )
        {
        outputPoint[j] += static_cast<ScalarType>( 
          weights[counter] * m_Iterator[j].Get());
        }

      // populate the indices array
      indices[counter] = &(m_Iterator[0].Value()) - basePointer;

      // go to next coefficient in the support region
      ++ counter;
      for ( j = 0; j < SpaceDimension; j++ )
        {
        ++( m_Iterator[j] );
        }
      }
  
    // return results
    for ( j = 0; j < SpaceDimension; j++ )
      {
      outputPoint[j] += transformedPoint[j];
      }

    }
    else
    {

    itkWarningMacro( << "B-spline coefficients have not been set" );

    for ( j = 0; j < SpaceDimension; j++ )
      {
      outputPoint[j] = transformedPoint[j];
      }

    }

}


// Transform a point
template<class TScalarType, unsigned int NDimensions, unsigned int VSplineOrder>
typename ADRevolveBSplineDeformableTransform<TScalarType, NDimensions,VSplineOrder>
::OutputPointType
ADRevolveBSplineDeformableTransform<TScalarType, NDimensions,VSplineOrder>
::TransformPoint(const InputPointType &point) const 
{
  
  WeightsType weights( m_WeightsFunction->GetNumberOfWeights() );
  ParameterIndexArrayType indices( m_WeightsFunction->GetNumberOfWeights() );
  OutputPointType outputPoint;
  bool inside;

  this->TransformPoint( point, outputPoint, weights, indices, inside );

  return outputPoint;

}
*/


// EGK: The following modified function replaces the previous two.
// Transform a point
template<class TScalarType, unsigned int NDimensions, unsigned int VSplineOrder>
typename ADRevolveBSplineDeformableTransform<TScalarType, NDimensions,VSplineOrder>
::OutputPointType
ADRevolveBSplineDeformableTransform<TScalarType, NDimensions,VSplineOrder>
::TransformPoint(const InputPointType &point) const 
{
  
  WeightsType weights( m_WeightsFunction->GetNumberOfWeights() );
//  ParameterIndexArrayType indices( m_WeightsFunction->GetNumberOfWeights() );
  OutputPointType outputPoint;
  bool inside;

//  this->TransformPoint( point, outputPoint, weights, indices, inside );

  unsigned int j;
  IndexType supportIndex;

//  InputPointType transformedPoint;
  if ( m_BulkTransform )
    {
//    transformedPoint = m_BulkTransform->TransformPoint( point );
    outputPoint = m_BulkTransform->TransformPoint( point );
    }
  else
    {
//    transformedPoint = point;
    outputPoint = point;
    }

  if ( m_CoefficientImage[0] )
    {

    ContinuousIndexType index;
    for ( j = 0; j < SpaceDimension; j++ )
      {
      index[j] = ( point[j] - m_GridOrigin[j] ) / m_GridSpacing[j];
      }

    // NOTE: if the support region does not lie totally within the grid
    // we assume zero displacement and return the input point
    inside = this->InsideValidRegion( index );
    if ( !inside )
      {
      return outputPoint;
//      outputPoint = transformedPoint;
//      return;
      }

    // Compute interpolation weights
    m_WeightsFunction->Evaluate( index, weights, supportIndex );

    // For each dimension, correlate coefficient with weights
    RegionType supportRegion;
    supportRegion.SetSize( m_SupportSize );
    supportRegion.SetIndex( supportIndex );

//    outputPoint.Fill( itk::NumericTraits<ScalarType>::Zero );

    typedef itk::ImageRegionConstIterator<ImageType> IteratorType;
    IteratorType m_Iterator[ SpaceDimension ];
    unsigned long counter = 0;
//    const PixelType * basePointer = m_CoefficientImage[0]->GetBufferPointer();

    for ( j = 0; j < SpaceDimension; j++ )
      {
      m_Iterator[j] = IteratorType( m_CoefficientImage[j], supportRegion );
      }

    while ( ! m_Iterator[0].IsAtEnd() )
      {

      // multiply weight with coefficient
      for ( j = 0; j < SpaceDimension; j++ )
        {
        outputPoint[j] += static_cast<ScalarType>( 
          weights[counter] * m_Iterator[j].Get());
        }

      // populate the indices array
//      indices[counter] = &(m_Iterator[0].Value()) - basePointer;

      // go to next coefficient in the support region
      ++ counter;
      for ( j = 0; j < SpaceDimension; j++ )
        {
        ++( m_Iterator[j] );
        }
      }
  
    // return results
//    for ( j = 0; j < SpaceDimension; j++ )
//      {
//      outputPoint[j] += transformedPoint[j];
//      }

    }
    else
    {

    itkExceptionMacro( << "B-spline coefficients have not been set" );
//    itkWarningMacro( << "B-spline coefficients have not been set" );

//    for ( j = 0; j < SpaceDimension; j++ )
//      {
//      outputPoint[j] = transformedPoint[j];
//      }

    }

  return outputPoint;

}







// Transform a point. Modified to allow for automatic differentiation.
template< class TScalarType, unsigned int NDimensions, unsigned int VSplineOrder>
void
ADRevolveBSplineDeformableTransform< TScalarType, NDimensions, VSplineOrder>
::TransformPointAD( const InputPointType& point, 
		    adoublev& outputPointAD ) const
{

  WeightsType weights( m_WeightsFunction->GetNumberOfWeights() );
//EGK: indices not needed
//  ParameterIndexArrayType indices( m_WeightsFunction->GetNumberOfWeights() );
//  OutputPointType outputPoint;
  bool inside;

     

   unsigned int j;
   IndexType supportIndex;

   if ( s_BulkTransform )
   {
      InputPointType transformedPoint = s_BulkTransform->TransformPoint( point );

//      for ( j = 0; j < SpaceDimension; j++ )
//      {
//	 this->m_TransformedPointAD[j] = transformedPoint[j];
//      }
      outputPointAD = transformedPoint.GetDataPointer();

   }
   else
   {
//      for ( j = 0; j < SpaceDimension; j++ )
//      {
//	 this->m_TransformedPointAD[j] = point[j];
//      }
      outputPointAD = const_cast<double*>(point.GetDataPointer());
   }

   if ( s_CoefficientImage[0] )
   {

//      unsigned int numberOfParametersPerDimension = 
//	 this->GetNumberOfParametersPerDimension();

      ContinuousIndexType index;
      for ( j = 0; j < SpaceDimension; j++ )
      {
	 index[j] = ( point[j] - s_GridOrigin[j] ) / s_GridSpacing[j];
      }

      // NOTE: if the support region does not lie totally within the grid
      // we assume zero displacement and return the input point
      inside = this->InsideValidRegionAD( index );
      if ( !inside )
      {
//	 return this->m_TransformedPointAD;
//	 outputPointAD = this->m_TransformedPointAD;
	 return;
      }

      // Compute interpolation weights
      m_WeightsFunction->Evaluate( index, weights, supportIndex );

      // For each dimension, correlate coefficient with weights
      RegionType supportRegion;
      supportRegion.SetSize( m_WeightsFunction->GetSupportSize() );
      supportRegion.SetIndex( supportIndex );

//      outputPointAD = 0.0;

      typedef itk::ImageRegionConstIterator<ImageType> IteratorType;
      IteratorType m_Iterator[ SpaceDimension ];
      unsigned long counter = 0;
      const PixelType * basePointer = s_CoefficientImage[0]->GetBufferPointer();

      for ( j = 0; j < SpaceDimension; j++ )
      {
	 m_Iterator[j] = IteratorType( s_CoefficientImage[j], supportRegion );
      }

      while ( ! m_Iterator[0].IsAtEnd() )
      {

	 // multiply weight with coefficient
	 for ( j = 0; j < SpaceDimension; j++ )
	 {

	    outputPointAD[j] += weights[counter] *
	       s_InputParametersPointerAD[ &(m_Iterator[j].Value()) - 
					   basePointer ];
	    
	 }

	 // populate the indices array (EGK: Not needed)
//	 indices[counter] = &(m_Iterator[0].Value()) - basePointer;

	 // go to next coefficient in the support region
	 ++counter;
	 for ( j = 0; j < SpaceDimension; j++ )
	 {
	    ++( m_Iterator[j] );
	 }
      }
  
      // return results
//      for ( j = 0; j < SpaceDimension; j++ )
//      {
//	 outputPointAD[j] += this->m_TransformedPointAD[j];
//      }
//      outputPointAD += this->m_TransformedPointAD;

   }
   else
   {

      //itkWarningMacro( << "B-spline coefficients have not been set" );
      itkExceptionMacro( << "B-spline coefficients have not been set" );

//      for ( j = 0; j < SpaceDimension; j++ )
//      {
//	 outputPointAD[j] = this->m_TransformedPointAD[j];
//      }
//      outputPointAD = this->m_TransformedPointAD;

   }


}




 
// Compute the Jacobian in one position 
template<class TScalarType, unsigned int NDimensions, unsigned int VSplineOrder>
const 
typename ADRevolveBSplineDeformableTransform<TScalarType, NDimensions,VSplineOrder>
::JacobianType & 
ADRevolveBSplineDeformableTransform<TScalarType, NDimensions,VSplineOrder>
::GetJacobian( const InputPointType & point ) const
{

  // Zero all components of jacobian
  // NOTE: for efficiency, we only need to zero out the coefficients
  // that got fill last time this method was called.
  RegionType supportRegion;
  supportRegion.SetSize( m_SupportSize );
  supportRegion.SetIndex( m_LastJacobianIndex );

  typedef itk::ImageRegionIterator<JacobianImageType> IteratorType;
  IteratorType m_Iterator[ SpaceDimension ];
  unsigned int j;

  for ( j = 0; j < SpaceDimension; j++ )
    {
    m_Iterator[j] = IteratorType( m_JacobianImage[j], supportRegion );
    }

  while ( ! m_Iterator[0].IsAtEnd() )
    {

    // zero out jacobian elements
    for ( j = 0; j < SpaceDimension; j++ )
      {
      m_Iterator[j].Set( itk::NumericTraits<JacobianPixelType>::Zero );
      }

    for ( j = 0; j < SpaceDimension; j++ )
      {
      ++( m_Iterator[j] );
      }
    }

 
  ContinuousIndexType index;
  for ( j = 0; j < SpaceDimension; j++ )
    {
    index[j] = ( point[j] - m_GridOrigin[j] ) / m_GridSpacing[j];
    }

  // NOTE: if the support region does not lie totally within the grid
  // we assume zero displacement and return the input point
  if ( !this->InsideValidRegion( index ) )
    {
    return this->m_Jacobian;
    }

  // Compute interpolation weights
  WeightsType weights( m_WeightsFunction->GetNumberOfWeights() );
  IndexType supportIndex;

  m_WeightsFunction->Evaluate( index, weights, supportIndex );
  m_LastJacobianIndex = supportIndex;

  // For each dimension, copy the weight to the support region
  supportRegion.SetIndex( supportIndex );
  unsigned long counter = 0;

  for ( j = 0; j < SpaceDimension; j++ )
    {
    m_Iterator[j] = IteratorType( m_JacobianImage[j], supportRegion );
    }

  while ( ! m_Iterator[0].IsAtEnd() )
    {

    // copy weight to jacobian image
    for ( j = 0; j < SpaceDimension; j++ )
      {
      m_Iterator[j].Set( static_cast<JacobianPixelType>( weights[counter] ) );
      }

    // go to next coefficient in the support region
    ++ counter;
    for ( j = 0; j < SpaceDimension; j++ )
      {
      ++( m_Iterator[j] );
      }
    }


  // Return the results
  return this->m_Jacobian;

}






// The transform is saved in the following format:
// where # means a number
/*
NameOfClass
spacing # # #
origin # # #
size # # #
index # # #
affine
# # # #
# # # #
# # # #
displacements
# # # ...
... 
...


*/


template<class TScalarType, unsigned int NDimensions, unsigned int VSplineOrder>
void
ADRevolveBSplineDeformableTransform<TScalarType, NDimensions,VSplineOrder>
::SaveTransform(const char* filename) const
{


   std::ofstream fout(filename);
   if(!fout)
   {
      itkExceptionMacro(<< "Error: could not open file for saving transform");
   }

   fout.setf(std::ios_base::scientific,std::ios_base::floatfield);
   fout.precision(16);

   typename RegionType::SizeType size = m_GridRegion.GetSize();
   typename RegionType::IndexType index = m_GridRegion.GetIndex();

   fout << this->GetNameOfClass() << "\n";
   fout << "spacing " 
	<< m_GridSpacing[0] << " " 
	<< m_GridSpacing[1] << " " 
	<< m_GridSpacing[2] << "\n";
   fout << "origin " 
	<< m_GridOrigin[0] << " " 
	<< m_GridOrigin[1] << " " 
	<< m_GridOrigin[2] << "\n";
   fout << "size " << size[0] << " " << size[1] << " " << size[2] << "\n";
   fout << "index " << index[0] << " " << index[1] << " " << index[2] << "\n";

   fout << "affine " << "\n";

   // Cast the bulk transform to an affine transform. If it's not an affine
   // transform assume identity.
   typedef itk::AffineTransform<ScalarType,3> AffineTransformType;

   const AffineTransformType* affine = 
      dynamic_cast< const AffineTransformType*>(this->GetBulkTransform());
   
   if(affine==0)
   {

      // Assume identity
      fout << "1 0 0 0\n0 1 0 0\n0 0 1 0\n";

   }
   else
   {
   for(int i=0; i < 3; i++)
   {
      for(int j=0; j < 3; j++)
      {
	 fout << affine->GetMatrix()[i][j] << "  ";
      }
      fout << affine->GetOffset()[i] << "\n";
   }
   }


   fout << "displacements";

   const ParametersType& parameters = this->GetParameters();

   int n = this->GetNumberOfParameters();
   
   for(int i = 0;i<n;i++)
   {
      if(i%3==0)
	 fout << "\n";
      fout << parameters[i] << " ";
   }

   fout << "\n";

   fout.close();
}





template<class TScalarType, unsigned int NDimensions, unsigned int VSplineOrder>
void
ADRevolveBSplineDeformableTransform<TScalarType, NDimensions,VSplineOrder>
::LoadTransform(const char* filename)
{

   
   std::ifstream fin(filename);

   if(!fin)
   {
      itkExceptionMacro(<< "Error: could not open file for loading transform");
   }

   std::string words;

   fin >> words;
   if(words != this->GetNameOfClass() )
   {
      itkExceptionMacro(<< "Error reading input file " << words << " ");
   }


   RegionType region;
   typename RegionType::SizeType size;
   typename RegionType::IndexType index;
   SpacingType spacing;
   OriginType origin;
   

   fin >> words >> spacing[0] >> spacing[1] >> spacing[2];
   fin >> words >> origin[0] >> origin[1] >> origin[2];
   fin >> words >> size[0] >> size[1] >> size[2];
   fin >> words >> index[0] >> index[1] >> index[2];

   region.SetSize(size);
   region.SetIndex(index);

   typedef itk::AffineTransform<ScalarType,3> AffineTransformType;

   typename AffineTransformType::Pointer affineTransform = AffineTransformType::New();

   typename AffineTransformType::ParametersType affineParameters(12);


   fin >> words;

   for(int i=0;i<12;i++)
   {
      fin >> affineParameters[i];
   }


   typename AffineTransformType::MatrixType matrix;
   typename AffineTransformType::OffsetType offset;

   int c=0;
   for(int i=0; i < 3; i++)
   {
      for(int j=0; j < 3; j++)
       {
	 matrix[i][j] = affineParameters[c++];
      }
      offset[i] = affineParameters[c++];
   }

   
   affineTransform->SetMatrix(matrix);
   affineTransform->SetOffset(offset);



   unsigned int numberOfParameters = 3*size[0]*size[1]*size[2];

   // Normally the bspline transform's parameters are managed by the
   // caller. Since we're loading the bspline from file, we, the bspline
   // object itself, is the caller. Hence we use the m_Parameters array from
   // one of the base classes.
   this->m_Parameters = ParametersType( numberOfParameters );


   fin >> words;

   for(unsigned int i=0;i<numberOfParameters;i++)
   {
      fin >> this->m_Parameters[i];
   }


   fin.close();


   this->SetGridSpacing( spacing );
   this->SetGridOrigin( origin );
   this->SetGridRegion( region );
   this->SetBulkTransform( affineTransform );
   this->SetParameters( this->m_Parameters );


}
 
 
  
} // namespace itkek

#endif /* _itkekADRevolveBSplineDeformableTransform_txx */
